;;;
;;; StraightenAndCrop.scm
;;;
;;; Copyright (C) 2009 Richard McLean (programmer_ceds A T yahoo D O T co D O T uk)
;;;

;;; This program is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Notes for the Straighten and Crop GIMP script

;;; 1.	Before running the script select the "Paths" tool from the toolbox. Ensure that the "Edit Mode" is set to "Design".
;;; 	The image must contain no paths at this stage.
;;; 
;;; 2.	Left-click at one point on the line that is to be made horizontal (or vertical).
;;; 
;;; 3.	Left-click at a second point on the same line (use separate clicks to define the two points; do not hold the left
;;; 	mouse button and drag - this will not generate a path with the required two end (or anchor) points)
;;; 
;;; 4.	Use the left mouse button to drag the two end-points of the path so that it lies over, or parallel to, the line that is
;;; 	to become horizontal (or vertical). Bending the path into a curve may make it harder to determine its correct position
;;; 	but the script will	still rotate the image based on the positions of the end-points of the path.
;;; 
;;; 5.	To make a path that has become invisible visible again use the Paths dialog (not the Paths tool) and right-click on the
;;; 	name of the path and select the "Path Tool" option. The Paths dialog can be docked in the toolbox - left-click the
;;; 	"Configure this tab" icon (the left pointing arrow in a square icon to the right of the toolbox tab description) and
;;; 	select "Add Tab/Paths". Alternatively display the dialog using "Windows/Dockable Dialogs/Paths".
;;; 
;;; 6.	Select "Tools/Transform Tools/Straighten and Crop"
;;; 
;;; 7.	In the "Auto" mode a path that is at an angle of 45 degrees or less to the horizontal will be taken to be the desired
;;; 	horizontal. Paths at angles greater than 45 degrees will be taken to be the desired vertical.
;;; 
;;; 8.	The "Crop after rotating" option allows for the optional cropping of the image to remove the triangular
;;;		transparent areas that are introduced by the rotation, leaving an opaque rectangular image. If either of the GIMP
;;;		resize options are chosen the "GIMP resize type" option allows for the selection of the resizing to be performed by the
;;;		internal GIMP software. If the "Post Rotate Crop" option is selected the three check boxes that apply
;;;		to this script should be set as required.
;;;
;;;		Note that if the cropping is performed by The GIMP this script removes transparent pixels that The GIMP
;;;		can leave in the corners.
;;;
;;;		The option to rotate but not crop the image is also provided.

;;; 19.7.2010
;;;		a) the undo group is ended before calling Post Rotate Crop to prevent a GIMP warning about
;;;		   an inconsistent state if Post Rotate Crop throws an error
;;;		b) the specified aspect ratio height and width are limited to a minimum value of 1
;;;
;;; 14.12.2010
;;;     The script can now use either the Script-Fu or Python-Fu version of the Post Rotate Crop script. By
;;;     preference it uses the Python version.
;;;

(define (script-fu-straighten-and-crop inImage inLayer inMode inCentre inX inY inInterpolation inDelPath
		 inCropType inResizeType inPostRotateCropMaximizeType inPostRotateCropAspectWidth inPostRotateCropAspectHeight)
	(let* (
        (newpath 0)
        (strokes 0)
        (num_vectors 0)
        (count 0)
        (num_points 0)
        (points 0)
        (point1_x 0)
        (point1_y 0)
        (point2_x 0)
        (point2_y 0)
        (left_x 0)
        (left_y 0)
        (right_x 0)
        (right_y 0)
        (x_difference 0)
        (y_difference 0)
        (direction 0)
        (angle 0.0)
        (line_centre_x 0)
        (line_centre_y 0)
        (rotate_about_image_centre FALSE)
        (resize_type 0)
        (change_made 0)
        (top_left_corner 0)
        (top_right_corner 0)
        (bottom_left_corner 0)
        (bottom_right_corner 0)
        (top_edge 0)
        (bottom_edge 0)
        (left_edge 0)
        (right_edge 0)
        (width 0)
        (height 0)
		(pixel_value '())
		(x_offset 0)
		(y_offset 0)
 		(alpha 3)

        (major_version_no 0)
        (minor_version_no 0)
        (release_no 0)
        (version_list (strbreakup (car (gimp-version)) "."))

; uncomment the following lines if want to check that the path is straight
 ;       (control_1a_x 0)
 ;       (control_1a_y 0)
 ;       (control_1b_x 0)
 ;       (control_1b_y 0)
 ;       (control_2a_x 0)
 ;       (control_2a_y 0)
 ;       (control_2b_x 0)
 ;       (control_2b_y 0)
		)
		
		(define MODE_AUTO       0)
		(define MODE_HORIZONTAL 1)
		(define MODE_VERTICAL   2)
		
		(define CENTRE_IMAGE     0)
		(define CENTRE_PATH      1)
		(define CENTRE_SPECIFIED 2)

		(define POST_ROTATE_CROP   0)
		(define GIMP_PLUS_AUTOCROP 1)
		(define GIMP_NO_AUTOCROP   2)
		(define NO_CROP            3)

  (set! major_version_no (string->number (car version_list) 10))
  (set! minor_version_no (string->number (cadr version_list) 10))
  (set! release_no (string->number (caddr version_list) 10))
  
  (if (or (< major_version_no 2)
  		  (and (= major_version_no 2) (< minor_version_no 6))
  		  (and (= major_version_no 2) (= minor_version_no 6) (< release_no 7)))
  	(error '(The rotate can fail with GIMP versions before V2.6.7 - please update GIMP from ) (car (gimp-version)))
  )	; end - if
	
  (gimp-image-undo-group-start inImage)
  (gimp-selection-none inImage)

  (if (and (= inCropType POST_ROTATE_CROP) (= (car (gimp-procedural-db-proc-exists "script-fu-post-rotate-crop")) FALSE)
               (= (car (gimp-procedural-db-proc-exists "python-fu-Post-Rotate-Crop")) FALSE))
    (prog1
      (gimp-image-undo-group-end inImage)  
      (gimp-message "Script Post Rotate Crop (file PostRotateCrop.scm or PostRotateCrop.py) not found")
      (quit)
    )	; end - prog1
  )	; end - if

  (set! num_vectors (car (gimp-image-get-vectors inImage))) 
  (if (not (= num_vectors 1))
    (prog1
      (gimp-image-undo-group-end inImage)  
      (error '(The drawable must have a single straight line path defined for the script to work) num_vectors)
    )	; end - prog1
  )	; end - if

  (set! newpath (vector-ref (cadr (gimp-image-get-vectors inImage)) 0)) 
   
  (set! strokes (car (gimp-vectors-get-strokes newpath)))
  (if (not (= strokes 1))
    (prog1
      (gimp-image-undo-group-end inImage)  
      (error '(The single path must be a straight line) strokes)
    )	; end - prog1
  )	; end - if

  (set! num_points(cadr (gimp-vectors-stroke-get-points newpath 1)))
  (if (not(= num_points 12))	; (there are three coordinate pairs per point)
    (prog1
      (gimp-image-undo-group-end inImage)  
      (error '(The path must contain exactly two points) (/ num_points 6))
    )	; end - prog1
  )	; end - if

  (set! points (caddr (gimp-vectors-stroke-get-points newpath 1)))

  ; read the anchor points at the ends of the line
  ;
  (set! point1_x (vector-ref points 2))
  (set! point1_y (vector-ref points 3))
  (set! point2_x (vector-ref points 8))
  (set! point2_y (vector-ref points 9))

; uncomment the following lines if want to check that the path is straight
;  (set! control_1a_x (vector-ref points 0))
;  (set! control_1a_y (vector-ref points 1))
;  (set! control_1b_x (vector-ref points 4))
;  (set! control_1b_y (vector-ref points 5))
;  (set! control_2a_x (vector-ref points 6))
;  (set! control_2a_y (vector-ref points 7))
;  (set! control_2b_x (vector-ref points 10))
;  (set! control_2b_y (vector-ref points 11))
;  
;  (if (not (and (= point1_x control_1a_x)
;                (= point1_x control_1b_x)
;                (= point1_y control_1a_y)
;                (= point1_y control_1b_y)
;                (= point2_x control_2a_x)
;                (= point2_x control_2b_x)
;                (= point2_y control_2a_y)
;                (= point2_y control_2b_y)))
;    (gimp-message "Using the endpoints of a curved path")
;  )	; end - if
  
  
  (if (< point1_x point2_x)
    (prog1
      (set! left_x point1_x)
      (set! left_y point1_y)
      (set! right_x point2_x)
      (set! right_y point2_y)
    )	; - end prog1
    (prog1
      (set! left_x point2_x)
      (set! left_y point2_y)
      (set! right_x point1_x)
      (set! right_y point1_y)
    )	; - end prog1
  )	; end - if
  
  (set! x_difference (- right_x left_x))
  
  (set! line_centre_x (+ left_x (/ (- right_x left_x) 2)))
  
  (if (> left_y right_y)
    (prog1
      (set! y_difference (- left_y right_y))
      (set! direction 0)		; rotate forwards
      (set! line_centre_y (+ right_y (/ (- left_y right_y) 2)))
    )	; end - prog1
    (prog1
      (set! y_difference (- right_y left_y))
      (set! direction 1)		; rotate backwards
      (set! line_centre_y (+ left_y (/ (- right_y left_y) 2)))
    )	; end - prog1
  )	; end - if
  
  (if (or (= inMode MODE_HORIZONTAL) (and (= inMode MODE_AUTO) (>= x_difference y_difference)))
    (prog1								; make the path horizontal
      (if (= x_difference 0)
        (set! angle (/ *pi* 2))	; (90 degrees as radians)
        (set! angle (atan (/ y_difference x_difference)))
      )	; end - if
    )	; end - prog1
    (prog1								; make the path vertical
      (if (= y_difference 0)
        (set! angle (/ *pi* 2))	; (90 degrees as radians)
        (set! angle (atan (/ x_difference y_difference)))
      )	; end - if
      (set! direction (- 1 direction))	; invert the direction
    )	; end - prog1
  )	; end - if

  (if (= inCentre CENTRE_IMAGE)
    (set! rotate_about_image_centre TRUE)
    (set! rotate_about_image_centre FALSE)
  )	; end - if
  
  (if (= inCentre CENTRE_SPECIFIED)
    (prog1
    (set! line_centre_x inX)
    (set! line_centre_y inY)
    )	; end - prog1
  )	; end - if

  (if (or (= inCropType POST_ROTATE_CROP) (= inCropType NO_CROP))
    (set! resize_type TRANSFORM-RESIZE-ADJUST)	; no cropping when no crop req'd or going to use Post Rotate Crop
    (set! resize_type inResizeType)
  )	; end - if
  
  (gimp-drawable-transform-rotate inLayer angle rotate_about_image_centre line_centre_x line_centre_y direction
  				 inInterpolation TRUE 3 resize_type)

  (if (= inCropType GIMP_PLUS_AUTOCROP)
    (prog1
  	  (plug-in-autocrop RUN-NONINTERACTIVE inImage inLayer)
  	  										; remove transparent pixels left by the GIMP cropping
  	  (plug-in-threshold-alpha RUN-NONINTERACTIVE inImage inLayer 0)
      (set! change_made 1)
      (while (= change_made 1)
        (set! change_made 0)
        (set! width (car (gimp-image-width inImage)))
      	(set! height (car (gimp-image-height inImage)))
      	
        (set! top_left_corner 0)
        (set! top_right_corner 0)
        (set! bottom_left_corner 0)
        (set! bottom_right_corner 0)
        (set! top_edge 0)
        (set! bottom_edge 0)
        (set! left_edge 0)
        (set! right_edge 0)
        
        (set! x_offset 0)
        (set! y_offset 0)

        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer 0 0)))
        (if (= (aref pixel_value alpha) 0)
          (set! top_left_corner 1)
        )	; end - if
        
        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer 0 (- height 1))))
        (if (= (aref pixel_value alpha) 0)
          (set! bottom_left_corner 1)
        )	; end - if
        
        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer (- width 1) 0)))
        (if (= (aref pixel_value alpha) 0)
          (set! top_right_corner 1)
        )	; end - if
        
        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer (- width 1) (- height 1))))
        (if (= (aref pixel_value alpha) 0)
          (set! bottom_right_corner 1)
        )	; end - if
        

        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer 1 0)))
        (if (= (aref pixel_value alpha) 0)
          (set! top_edge 1)
        )	; end - if
        
        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer (- width 2) 0)))
        (if (= (aref pixel_value alpha) 0)
          (set! top_edge (+ top_edge 1))
        )	; end - if
        

        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer 1 (- height 1))))
        (if (= (aref pixel_value alpha) 0)
          (set! bottom_edge 1)
        )	; end - if
       
        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer (- width 2) (- height 1))))
        (if (= (aref pixel_value alpha) 0)
          (set! bottom_edge (+ bottom_edge 1))
        )	; end - if
        

        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer 0 1)))
        (if (= (aref pixel_value alpha) 0)
          (set! left_edge 1)
        )	; end - if
        
        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer 0 (- height 2))))
        (if (= (aref pixel_value alpha) 0)
          (set! left_edge (+ left_edge 1))
        )	; end - if
        

        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer (- width 1) 1)))
        (if (= (aref pixel_value alpha) 0)
          (set! right_edge 1)
        )	; end - if
        
        (set! pixel_value (cadr (gimp-drawable-get-pixel inLayer (- width 1) (- height 2))))
        (if (= (aref pixel_value alpha) 0)
          (set! right_edge (+ right_edge 1))
        )	; end - if
          
        
        (if (> top_edge 0)
          (prog1						; crop off the top row
            (set! change_made 1)
            (set! y_offset 1)
            (set! height (- height 1))
            (set! top_left_corner 0)
            (set! top_right_corner 0)
          )		; end - prog1
        )	; end - if
        
        (if (> bottom_edge 0)
          (prog1						; crop off the bottom row
            (set! change_made 1)
            (set! height (- height 1))
            (set! bottom_left_corner 0)
            (set! bottom_right_corner 0)
          )		; end - prog1
        )	; end - if
        
        (if (> left_edge 0)
          (prog1						; crop off the left column
            (set! change_made 1)
            (set! x_offset 1)
            (set! width (- width 1))
            (set! top_left_corner 0)
            (set! bottom_left_corner 0)
          )		; end - prog1
        )	; end - if
        
        (if (> right_edge 0)
          (prog1						; crop off the right column
            (set! change_made 1)
            (set! width (- width 1))
            (set! top_right_corner 0)
            (set! bottom_right_corner 0)
          )		; end - prog1
        )	; end - if
        
        (if (> width height)
          (prog1
            (if (> top_left_corner 0)
              (prog1					; crop off the left column
                (set! change_made 1)
                (set! x_offset 1)
                (set! width (- width 1))
                (set! bottom_left_corner 0)
              )	; end - prog1
            )	; end - if

            (if (> bottom_left_corner 0)
              (prog1					; crop off the left column
                (set! change_made 1)
                (set! x_offset 1)
                (set! width (- width 1))
;               (set! top_left_corner 0)
              )	; end - prog1
            )	; end - if

            (if (> top_right_corner 0)
              (prog1					; crop off the right column
                (set! change_made 1)
                (set! width (- width 1))
                (set! bottom_right_corner 0)
              )	; end - prog1
            )	; end - if

            (if (> bottom_right_corner 0)
              (prog1					; crop off the right column
                (set! change_made 1)
                (set! width (- width 1))
;               (set! top_right_corner 0)
              )	; end - prog1
            )	; end - if
          )		; end - prog1
          
          
          
          (prog1
            (if (> top_left_corner 0)
              (prog1					; crop off the top row
                (set! change_made 1)
                (set! y_offset 1)
                (set! height (- height 1))
                (set! top_right_corner 0)
              )	; end - prog1
            )	; end - if

            (if (> top_right_corner 0)
              (prog1					; crop off the top row
                (set! change_made 1)
                (set! y_offset 1)
                (set! height (- height 1))
;               (set! top_left_corner 0)
              )	; end - prog1
            )	; end - if

            (if (> bottom_left_corner 0)
              (prog1					; crop off the bottom row
                (set! change_made 1)
                (set! height (- height 1))
                (set! bottom_right_corner 0)
              )	; end - prog1
            )	; end - if

            (if (> bottom_right_corner 0)
              (prog1					; crop off the bottom row
                (set! change_made 1)
                (set! height (- height 1))
;               (set! bottom_left_corner 0)
              )	; end - prog1
            )	; end - if
          )		; end - prog1
        )	; end - if
        
       (gimp-image-crop inImage width height x_offset y_offset)		    
        
      )	; end - while
  	)	; end - prog1
  )	; end - if

  (if (= inDelPath TRUE)  
    (gimp-image-remove-vectors inImage newpath)	; delete the path
  )	; end - if
  
  (gimp-image-undo-group-end inImage)  

  (if (= inCropType POST_ROTATE_CROP)
    (if (= (car (gimp-procedural-db-proc-exists "python-fu-Post-Rotate-Crop")) TRUE)
      (python-fu-Post-Rotate-Crop FALSE inImage inLayer inPostRotateCropMaximizeType inPostRotateCropAspectWidth inPostRotateCropAspectHeight)
      (script-fu-post-rotate-crop inImage inLayer inPostRotateCropMaximizeType inPostRotateCropAspectWidth inPostRotateCropAspectHeight)
    ) ; end - if
  )	; end - if
  
  (gimp-displays-flush)
  ) ;;let
) ;;def

(script-fu-register
 "script-fu-straighten-and-crop"
 _"<Image>/Tools/Transform Tools/Straighten and Crop"
 "Rotate the image by the angle defined by a straight line two-point path, then, optionally, crop the image"
 "Richard McLean"
 "2009, Richard McLean (programmer_ceds A T yahoo D O T co D O T uk)"
 "24th September 2009"
 "RGB* GRAY* INDEXED*"
 SF-IMAGE "The Image" 0
 SF-DRAWABLE "The Layer" 0
 SF-OPTION     "Vertical/horizontal mode"       (list "Auto" "Make path horizontal" "Make path vertical")
 SF-OPTION     "Centre for rotation"            (list "Image centre" "Middle of path" "Centre specified below")
 SF-ADJUSTMENT "Rotation centre X" '(0 0 9999 1 10 0 1)
 SF-ADJUSTMENT "Rotation centre Y" '(0 0 9999 1 10 0 1)
 SF-ENUM       "Interpolation"                  '("InterpolationType" "cubic")
 SF-TOGGLE	   "Delete path after rotate" TRUE
 SF-OPTION     "Crop after rotating"            (list "Post Rotate Crop" "GIMP resize + Autocrop image" "GIMP resize" "No crop")
 SF-ENUM	   "GIMP resize type"               '("GimpTransformResize" "crop")
 SF-OPTION     "Post Rotate Crop maximize type" (list "Height and width" "Height" "Height at the expense of width" "Width" "Width at the expense of height"
 										 "Height, width and area" "Keep image aspect ratio" "Aspect ratio specified below"
										  "No maximization")
 SF-ADJUSTMENT "Aspect ratio width" '(3 1 9999 1 10 0 1)
 SF-ADJUSTMENT "Aspect ratio height" '(2 1 9999 1 10 0 1)
)
